         page    80,132
COMMENT  ;

Set of routines to DRAW a Hi-Res line on an Alternate screen, and to map that
screen to the physical screen.
         ;

Clipping =       0                ; 1 = Clipping on, 0 = Clipping off

         public  ClrAScrn
         public  SetPScrn
         public  draw

         assume  cs:code, ds:data, es: Screen

data     segment byte public

DRow     dw      ?                ; Delta row
DCol     dw      ?                ; Delta col
RowStep  dw      ?                ; Row increment
ColStep  dw      ?                ; Col increment

OfsTable label   word             ; row offset table
Ofs      =       0                ; counting var
OddDisp  =       2000h            ; odd scan lines are in seg ba00
         rept    100
         DW      Ofs              ;; even scan line
         DW      Ofs + OddDisp    ;; odd scan line
Ofs      =       Ofs + 80
         endm

data     ends


Screen   segment para

AltScrn  db      16384 dup (?)

Screen   ends


code     segment byte public

ClrAScrn proc    near             ; CLeaR Alternate SCReeN
         mov     ax,seg AltScrn
         mov     es,ax            ; Address AltScrn w/ ES
         xor     di,di            ; ES:DI is top left of screen
         mov     cx,16384/2       ; we'll do a word fill
         xor     ax,ax            ; fill with 0s
         cld                      ; make sure we INCREMENT di
         rep     stosw            ; fill alt screen w/ 0s
         ret
ClrAScrn endp

SetPScrn proc    near
         mov     dx,ds            ; movsb uses ds ...
         mov     ax,seg AltScrn
         mov     ds,ax            ; AltScrn is the source
         mov     ax,0b800h
         mov     es,ax            ; CGA is the destination
         xor     si,si            ; copying from top left ...
         xor     di,di            ; ... to top left
         mov     cx,16384/2       ; word copy mimimizes CPU overhead
         cld                      ; make sure we INCREMENT di, si
         rep     movsw            ; do the copy
         mov     ds,dx            ; restore dx
         ret
SetPScrn endp

if       Clipping
Clip     macro   Coord, Limit     ;; Is 0 <= Coord <= Limit?
         local   Good, Bad        ;; local labels
         test    Coord,8000H      ;; check the sign bit
         jnz     Bad              ;; Coord is negative
         cmp     Coord,Limit      ;; Compare to limit
         jbe     Good             ;; OK if <=
Bad:     ret                      ;; return to Draw without doing anything
Good:                             ;; keep going
         endm
endif

WriteDot proc    near             ; CX = Row, DX = Col, BP = Color

if       Clipping
         Clip    cx,199           ; 200 rows, 0..199
         Clip    dx,639           ; 640 cols, 0..639
endif
         push    bx
         push    cx
         mov     ax,si            ; preserve value
         mov     si,cx            ; we need ofstable[row]
         shl     si,1             ; offsets are words
         mov     bx,OfsTable[si]  ; offset of row start
         mov     si,ax            ; restore global si
         mov     ax,dx            ; col
         shr     ax,1             ; col/2
         shr     ax,1             ; col/4
         shr     ax,1             ; col/8  [eight pixels/byte]
         add     bx,ax            ; addr of appropriate sceen byte
         mov     cx,dx            ; col
         and     cl,7             ; pixel spec
         inc     cl               ; add 1, so ROL will make pixel rightmost
         mov     al,es:[bx]       ; get screen byte
         rol     al,cl            ; make selected pixel the rightmost
         and     al,11111110B     ; clear rightmost pixel
         or      ax,bp            ; set color (we don't care about ah)
         ror     al,cl            ; move pixels back
         mov     es:[bx],al       ; write it back
         pop     cx               ; Restore CX and BX
         pop     bx
         RET                      ; That's all folks...
WriteDot ENDP

COMMENT  ;

DRAW a line from (Row1,Col1) to (Row2,Col2).  Works in any graphics mode, with
the appropriate WriteDot routine.

Arguments are:
  Row1:   cx
  Col1:   dx
  Row2:   si
  Col2:   di
  Color:  bp

Trashes all ?X and ?I registers, plus BP and ES.  * Not re-entrant *

Draw() register usage:
  AX = (general purpose)
  BX = Error term
  CX = current Row
  DX = current Col
  SI = error Limit
  DI = Row2/Col2 (stepped co-ordinate limit)
  BP = Color
  ES = Screen segment
         ;

Draw     proc    near
                                  ; WriteDot setup
         mov     ax,seg AltScrn   ; (alternate) screen segment
         mov     es,ax            ; we'll be addressing the screen a lot...

Calc     macro   A1,A2,Delta,Step ; Arg1/2, DeltaArg, StepArg
         local   Write
                                  ; calc Delta and Step
         mov     bx,1             ; We'll NEGate this if we have to...
         mov     ax,A2            ; Get <Arg2> value
         sub     ax,A1            ; Arg2 - Arg1
         jge     Write            ; We need do no more if Delta is positive
         neg     bx               ; Step := -1
         neg     ax               ; Delta := Abs(Arg2 - Arg1)
Write:   mov     [Delta],ax       ; write values to mem
         mov     [Step],bx
         endm

         Calc    dx,di,DCol,ColStep
         Calc    cx,si,DRow,RowStep

                                  ; set error term & decide which dim. to step
         xor     bx,bx            ; Error := 0
         cmp     ax,[DCol]        ; We'll step the dimension that changes more
         jg      StepRows         ; We'll step the rows

Canonic  macro   ?Step,AltStep
         local   SetLimit         ; make sure we are stepping lo to hi
         mov     ax,[?Step]       ; ?Step is 1 if ?1 < ?2, ffff if not
         rcl     ax,1             ; set carry flag to sign bit
         jnc     SetLimit         ; exchange args if Col1 > Col2
         xchg    cx,si            ; exchange Row1, Row2
         xchg    dx,di            ; exchange Col1, Col2
         neg     [AltStep]
SetLimit:
         endm

StepCols:Canonic ColStep,RowStep  ; ELSE: We step the cols
         mov     si,[DCol]        ; read back

StepIt   macro   Coord,AltCoord,Delta,AltDelta,AltStep
         local   Top,Bottom
         sar     si,1             ;; ErrorLimit := Delta shr 1
Top:     cmp     Coord,di
         je      EndPlot          ;; WHILE Coord <> Lim DO ...
         call    WriteDot         ;; Write the dot at [CX, DX]
         add     bx,[AltDelta]    ;; Error := Error + AltDelta
         cmp     si,bx
         jge     Bottom           ;; if Error > ErrorLimit then ...
         sub     bx,[Delta]       ;; Error := Error - Delta
         add     AltCoord,[AltStep] ;; AltCoord := AltCoord + AltStep
Bottom:  inc     Coord            ;; Coord := Coord + 1
         jmp     Top              ;; end WHILE
         endm

         StepIt  dx,cx,DCol,DRow,RowStep ; step cols
                                         ; (macro contains JMP EndPlot)
StepRows:Canonic RowStep,ColStep  ; Step the rows
         mov     di,si            ; final row
         mov     si,[DRow]
         StepIt  cx,dx,DRow,DCol,ColStep ; step rows

EndPlot: ret
Draw     ENDP

CODE     ENDS
         END
