/* VESA package for emx/gcc --- Copyright (c) 1993 by Johannes Martin */
#include <sys/hw.h>
#include <stdlib.h>

#define INCL_VIO
#define INCL_KBD
#define INCL_DOSPROCESS
#define INCL_DOSFILEMGR

#define VESADLL

#include <os2emx.h>
#include <os2thunk.h>

#include "vgaports.h"
#include "vesa.h"
#include "vesadll.h"
#include "common.h"

int activeindex   = 0;
VESASWITCH swfunc = NULL;
int screenlocked  = 0;

void Error(char *s)
{
  ULONG dummy;

  DosWrite(2, s, strlen(s), &dummy);
}

VESABOOL OS2VesaGetInfo(PVESAINFO pVesaInfo)
{
  pVesaInfo->Version = 0x102;
  pVesaInfo->OEMName = "OS/2 VESA Driver by Johannes Martin";
  pVesaInfo->Modes   = Modes;
  return(TRUE);
}

VESABOOL OS2VesaGetModeInfo(VESAWORD Mode, PVESAMODEINFO pVesaModeInfo)
{
  int i;

  Mode &= 0x7fff;
  for (i = 0; Modes[i] != Mode; i++)
    if (Modes[i] == 0xFFFF)
      return(FALSE);
  *pVesaModeInfo = ModeInfos[i].Vesa;
  if ((i == activeindex) && OS2VesaLockScreen(0))
    {
      VESAWORD dummy1, dummy2;

      OS2VesaGetScanlineLength(&pVesaModeInfo->BytesPerScanline, &dummy1, &dummy2);
      pVesaModeInfo->WindowAStart = GetPhysBuf();
      pVesaModeInfo->WindowBStart = pVesaModeInfo->WindowAStart;
      OS2VesaUnlockScreen();
    }
  return(TRUE);
}

VESABOOL OS2VesaSetMode(VESAWORD Mode)
{
  int i;

  for (i = 0; Modes[i] != Mode & 0x7fff; i++)
    if (Modes[i] == 0xFFFF)
      return(FALSE);
  if (!SetMode(i))
    return(FALSE);
  activeindex = i;
  if (OS2VesaLockScreen(0))
    {
      SetSpecial(i);
      if (((Mode & 0x8000) == 0) && (ModeInfos[i].Vesa.ModeAttributes & 0x10))
        Clear();
      OS2VesaSetWindow(0, 0);
      OS2VesaSetWindow(1, 0);
      OS2VesaUnlockScreen();
    }
  return(TRUE);
}

VESABOOL OS2VesaGetMode(PVESAWORD Mode)
{
  *Mode = Modes[activeindex];
  return(TRUE);
}

VESABOOL OS2VesaSetScanlineLength(VESAWORD NumberOfPixels)
{
  int div, h, i;

  if (!OS2VesaLockScreen(0))
    return(FALSE);
  _outp8(CRT_I, 0x14);
  if (_inp8(CRT_D) & 0x40)
    div = 8;
  else
    {
      _outp8(CRT_I, 0x17);
      if (_inp8(CRT_D) & 0x40)
        div = 2;
      else
        div = 4;
    }
  _outp8(CRT_I, 0x13);
  h = NumberOfPixels * ModeInfos[activeindex].Vesa.NumberOfBitsPerPixel / 8;
  i = h / div;
  if (h % div != 0)
    i++;
  _outp8(CRT_D, i & 0xFF);
  OS2VesaUnlockScreen();
  return(TRUE);
}

VESABOOL OS2VesaGetScanlineLength(PVESAWORD BytesPerScanline,
                                  PVESAWORD NumberOfPixels,
                                  PVESAWORD NumberOfScanlines)
{
  int mult;

  if (!OS2VesaLockScreen(0))
    return(FALSE);
  _outp8(CRT_I, 0x14);
  if (_inp8(CRT_D) & 0x40)
    mult = 8;
  else
    {
      _outp8(CRT_I, 0x17);
      if (_inp8(CRT_D) & 0x40)
        mult = 2;
      else
        mult = 4;
    }
  _outp8(CRT_I, 0x13);
  if ((*BytesPerScanline = _inp8(CRT_D)) == 0)
    *BytesPerScanline = mult << 8;
  else
    *BytesPerScanline *= mult;
  *NumberOfPixels    = *BytesPerScanline * 8 / ModeInfos[activeindex].Vesa.NumberOfBitsPerPixel;
  *NumberOfScanlines = ModeInfos[activeindex].Vesa.Height;
  OS2VesaUnlockScreen();
  return(TRUE);
}

static VOID APIENTRY RedrawWaitThread(ULONG Param)
{
  static VESAWORD ScanlineLength = 0;
  static VESAWORD WindowAStat    = 0;
  static VESAWORD WindowBStat    = 0;

  USHORT notifytype;

  while (1)
    {
      VioSavRedrawWait(0, &notifytype, 0);
      screenlocked++;
      if (notifytype)
        {
          SetMode(activeindex);
          SetSpecial(activeindex);
          if (ScanlineLength != 0)
            OS2VesaSetScanlineLength(ScanlineLength);
          OS2VesaSetWindow(0, WindowAStat);
          OS2VesaSetWindow(1, WindowBStat);
        }
      swfunc(notifytype);
      if (!notifytype)
        {
          VESAWORD dummy1, dummy2;

          OS2VesaGetScanlineLength(&dummy1, &ScanlineLength, &dummy2);
          OS2VesaGetWindow(0, &WindowAStat);
          OS2VesaGetWindow(1, &WindowBStat);
          SetMode(0);
          SetSpecial(0);
        }
      screenlocked--;
    }
}

VESABOOL OS2VesaSetSwitchFunc(VESASWITCH SwitchFunc)
{
  static TID tid = 0;

  if ((VioSavRedrawUndo(1, 1, 0) != 0) && (tid != 0))
    DosKillThread(tid);
  if ((swfunc = SwitchFunc) != NULL)
    return(DosCreateThread(&tid, RedrawWaitThread, 0, 0, 32 * 1024) == 0);
  else
    {
      tid = 0;
      return(TRUE);
    }
}

VESABOOL OS2VesaLockScreen(VESACHAR wait)
{
  UCHAR Success;

  if ((screenlocked) ||
      ((VioScrLock(wait, &Success, 0) == 0) && (wait || (Success == 0))))
    {
      screenlocked++;
      return(TRUE);
    }
  else
    return(FALSE);
}

VESABOOL OS2VesaUnlockScreen(void)
{
  if (screenlocked)
    if (--screenlocked)
      return(TRUE);
    else
      return(VioScrUnLock(0) == 0);
  else
    return(FALSE);
}

VESABOOL OS2VesaSetModified(void)
{
  return(TRUE);
}

VESABOOL OS2VesaSetPalette(VESAWORD   StartIndex,
                           VESAWORD   NumberOfEntries,
                           PCVESACHAR Palette,
                           VESABOOL   Wait)
{
  if ((NumberOfEntries >= 1) && (StartIndex + NumberOfEntries <= 256))
    {
      if (!OS2VesaLockScreen(0))
        return(FALSE);
      if (Wait)
        _wait01 (IS1_R, 0x08);
      _outp8 (PEL_IW, StartIndex);
      _outps8dac (PEL_D, Palette, 3 * NumberOfEntries);
      OS2VesaUnlockScreen();
      return(TRUE);
    }
  return(FALSE);
}

VESABOOL OS2VesaGetPalette(VESAWORD  StartIndex,
                           VESAWORD  NumberOfEntries,
                           PVESACHAR Palette)
{
  if ((NumberOfEntries >= 1) && (StartIndex + NumberOfEntries <= 256))
    {
      int i, j;

      if (!OS2VesaLockScreen(0))
        return(FALSE);
      /* select palette register */
      _outp8(PEL_IR, StartIndex);

      for(j = 0; j < NumberOfEntries; j++)
        {
          for(i = 0; i < 10; i++) ;   /* delay (minimum 240ns) */
          *(Palette++) = (int) _inp8(PEL_D);
          for(i = 0; i < 10; i++) ;   /* delay (minimum 240ns) */
          *(Palette++) = (int) _inp8(PEL_D);
          for(i = 0; i < 10; i++) ;   /* delay (minimum 240ns) */
          *(Palette++) = (int) _inp8(PEL_D);
        }
      OS2VesaUnlockScreen();
      return(TRUE);
    }
  return(FALSE);
}

static ULONG LastCode = -1;

VESABOOL OS2VesaGetCharacter(PVESACHAR pchar)
{
  if (LastCode != -1)
    {
      *pchar   = LastCode;
      LastCode = -1;
    }
  else
    *pchar = _read_kbd(0, 1, 0);
  return(TRUE);
}

VESABOOL OS2VesaKeyboardHit(void)
{
  if (LastCode == -1)
    LastCode = _read_kbd(0, 0, 0);
  return(LastCode != -1);
}

int _CRT_init(void);
void _CRT_term(void);

unsigned long _DLL_InitTerm(unsigned long modhandle, unsigned long flag)
{
  if (flag)
    {
      if (activeindex)
        OS2VesaSetMode(0x03);
      ClearUp();
      _CRT_term();
      return(TRUE);
    }
  return((_CRT_init() == 0) && Initialize());
}
